/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.classifiers;

import de.unibi.techfak.scie.descriptors.OntologyReference;
import de.unibi.techfak.scie.descriptors.Paragraph;
import de.unibi.techfak.scie.descriptors.Sentence;
import de.unibi.techfak.scie.descriptors.Token;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

/**
 * This is a collection of static functions to calculate some standard features.
 *
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class FeatureUtils {

	public static double getRelativeIndex(int index, JCas jcas) {
		return (double) index / (double) jcas.getDocumentText().length();
	}

	public static double getRelativeBegin(Annotation anno, JCas jcas) {
		return (double) anno.getBegin() / (double) jcas.getDocumentText().length();
	}

	public static double getRelativeEnd(Annotation anno, JCas jcas) {
		return (double) anno.getEnd() / (double) jcas.getDocumentText().length();
	}

	public static double getRelativeLength(Annotation anno, JCas jcas) {
		return (double) (anno.getEnd() - anno.getBegin())
				/ (double) jcas.getDocumentText().length();
	}

	public static double getRelativeDist(Annotation a, Annotation b, JCas jcas) {
		final double dist;
		if (a.getEnd() < b.getBegin()) {
			dist = b.getBegin() - a.getEnd();
		} else if (b.getEnd() < a.getBegin()) {
			dist = a.getBegin() - b.getEnd();
		} else {
			dist = 0;
		}
		return dist / (double) jcas.getDocumentText().length();
	}

	/**
	 * Cleans the given input word for word identity features. This mainly
	 * means new lines.
	 *
	 * @param input the raw word.
	 *
	 * @return the cleaned word
	 */
	public static String cleanWord(final String input) {
		try {
			final StringBuilder out = new StringBuilder();
			final BufferedReader cleanR = new BufferedReader(new StringReader(input));
			String line;
			while ((line = cleanR.readLine()) != null) {
				out.append(line);
			}
			return out.toString();
		} catch (IOException ex) {
			throw new RuntimeException(ex);
		}
	}

	/**
	 * Calculates not the actual normal distribution N(input|0,sigma) /
	 * N(0|0,sigma)
	 *
	 * @param input the input for the density function.
	 * @param sigma the standard deviation of the normal distribution
	 *
	 * @return the value of N(input|0,sigma) / N(0|0,sigma)
	 */
	public static double calculateNormalDistribution(final double input,
			final double sigma) {
		final double normalizedExp = input / sigma;
		return Math.exp(-0.5 * normalizedExp * normalizedExp);
	}

	/**
	 * See below
	 */
	public static double calculateSyntacticProbability(Annotation a,
			Annotation b, final double sigma, final JCas jcas) {
		return calculateSyntacticProbability(a, b, sigma, 3, 10, jcas);
	}

	/**
	 * Calculates a probability (between 0 and 1) that models the confidence
	 * that two annotations belong together given their syntactic distance (the
	 * number of words, sentences and paragraphs in between).
	 *
	 * It is basically a gaussian-like function of that abovementioned distance.
	 *
	 * @param a first annotation.
	 * @param b second annotation.
	 *
	 * @return a probability (between 0 and 1) that models the confidence
	 * that two annotations belong together given their syntactic distance (the
	 * number of words, sentences and paragraphs in between).
	 */
	public static double calculateSyntacticProbability(Annotation a,
			Annotation b, final double sigma,
			final int sentenceMalus,
			final int paragraphMalus,
			final JCas jcas) {

		final int aBegin = a.getBegin();
		final int aEnd = a.getEnd();
		final int bBegin = b.getBegin();
		final int bEnd = b.getEnd();
		/*
		 * first treat the special cases that one of the annotations is covering
		 * the other or overlapping with it. If so we return 1.
		 */
		final Annotation first;
		final Annotation snd;

		if (aBegin < bBegin) {
			if (aEnd <= bBegin) {
				//a is before b
				first = a;
				snd = b;
			} else {
				//if both overlap we return 1
				return 1;
			}
		} else {
			if (bEnd <= aBegin) {
				//b is before a
				first = b;
				snd = a;
			} else {
				//if both overlap we return 1.
				return 1;
			}
		}
		/*
		 * get the words between both annotations. The number of those words is
		 * the first attempt at a distance.
		 */
		int distance = getDistanceInCoveringClass(Token.class, jcas, first, snd);
		/*
		 * If there are also Sentences in between, we punish that additionally.
		 */
		distance += sentenceMalus * getDistanceInCoveringClass(Sentence.class, jcas, first, snd);

		/*
		 * We do the same for paragraphs.
		 */
		distance += paragraphMalus * getDistanceInCoveringClass(Paragraph.class, jcas, first, snd);

		return calculateNormalDistribution(distance, sigma);
	}

	/**
	 * This calculates the distance of two annotations counted in the instance
	 * of the given Class.
	 * Please note that this method implicitly assumes that the annotations of
	 * the given class are
	 * directly following each other and that only one annotation of that class
	 * covers a given
	 * region at a time. This is only garantueed for primitive annotations of
	 * the document structure
	 * like Word, Sentence, Paragraph, etc.
	 *
	 * @param <X>
	 * @param annotationClass
	 * @param jcas
	 * @param first
	 * @param snd
	 *
	 * @return
	 */
	public static <X extends Annotation> int getDistanceInCoveringClass(
			Class<X> annotationClass, JCas jcas, Annotation first,
			Annotation snd) {
		final int firstBegin = first.getBegin();
		final int sndEnd = snd.getEnd();
		//get the number of annotations in the given class.
		final ArrayList<X> between = NonOverlappingCache.getInstance(jcas).selectBetween(annotationClass,
				firstBegin, sndEnd);
		if (between.isEmpty()) {
			return 0;
		} else {
			return between.size() - 1;
		}
	}

	public static final double getConfidence(
			final de.unibi.techfak.scie.descriptors.Annotation anno) {
		final double confidence;
		if (anno instanceof OntologyReference) {
			confidence = ((OntologyReference) anno).getProbability();
		} else {
			confidence = 1;
		}
		return confidence;
	}

	public static final int getDictionaryID(
			final de.unibi.techfak.scie.descriptors.Annotation anno) {
		if (anno instanceof OntologyReference) {
			return ((OntologyReference) anno).getId();
		} else {
			return 0;
		}
	}

	public static ArrayList<String> getPOSTags(Annotation anno, JCas jcas) {
		final ArrayList<String> out = new ArrayList<>();
		final List<Token> tokens = NonOverlappingCache.getInstance(jcas).
				selectBetween(Token.class, anno.getBegin(), anno.getEnd());
		for (final Token token : tokens) {
			out.add(token.getPOS());
		}
		return out;
	}

	public static ArrayList<String> getWordsInBetween(Annotation a,
			Annotation b, JCas jcas) {
		final ArrayList<String> out = new ArrayList<>();
		final int begin;
		final int end;
		if (a.getEnd() + 1 < b.getBegin()) {
			begin = a.getEnd() + 1;
			end = b.getBegin() - 1;
		} else if (b.getEnd() + 1 < a.getBegin()) {
			begin = b.getEnd() + 1;
			end = a.getBegin() - 1;
		} else {
			return out;
		}
		final List<Token> tokens = NonOverlappingCache.getInstance(jcas).
				selectBetween(Token.class, begin, end);
		for (final Token token : tokens) {
			out.add(token.getCoveredText());
		}
		return out;
	}
}
