/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.classifiers.data;

import java.util.SortedMap;
import java.util.TreeMap;

/**
 * This models represents features of a concrete DataPoint in terms of a map of
 * feature names to values to achieve a sparse feature representation in large
 * feature spaces.
 *
 * Note that this implies that all features that are not explicitly specified
 * are assumed to be zero.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class FeatureMap {

    private final TreeMap<Integer, Double> features = new TreeMap<>();
    private final FeatureDictionary dictionary;

    public FeatureMap(FeatureDictionary dictionary) {
        this.dictionary = dictionary;
    }

    public void addNumericFeature(final String name, final double value,
            boolean addFeaturesToDictionary) {
        if (!dictionary.hasFeature(name)) {
            if (addFeaturesToDictionary) {
                dictionary.addFeature(name);
            }
        }
        final int feature_idx = dictionary.getFeatureIndex(name);
        if (feature_idx != -1) {
            features.put(feature_idx, value);
        }
    }

    /**
     * This implicitly maps the given feature to the value 1.
     */
    public void addBooleanFeature(final String name,
            boolean addFeaturesToDictionary) {
        if (!dictionary.hasFeature(name)) {
            if (addFeaturesToDictionary) {
                dictionary.addFeature(name);
            }
        }
        final int feature_idx = dictionary.getFeatureIndex(name);
        if (feature_idx != -1) {
            features.put(feature_idx, 1.);
        }
    }

    /**
     * Returns the value for the given feature or null if the feature is
     * unknown.
     *
     * @param index
     * @return
     */
    public Double getValue(final int index) {
        return features.get(index);
    }

    /**
     * Returns the value for the given feature or null if the feature is
     * unknown.
     *
     * @param name
     * @return
     */
    public Double getValue(final String name) {
        final int index = dictionary.getFeatureIndex(name);
        if (index == -1) {
            return null;
        }
        return features.get(index);
    }

    public FeatureDictionary getDictionary() {
        return dictionary;
    }

    /**
     * Returns a sparse representation of the features for this datapoint using
     * the feature indices that are part of the FeatureDictionary.
     *
     * @return a sparse representation of the features for this datapoint using
     * the feature indices that are part of the FeatureDictionary.
     */
    public SortedMap<Integer, Double> getFeatures() {
        return features;
    }

}
